# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Detection of compute environments.

This module provides the :class:`ComputeEnvironment` class, which can be
subclassed to automatically detect specific computational environments.

This enables the user to adjust their workflow based on the present
environment, e.g. for the adjustment of scheduler submission scripts.
"""
import logging
import os
import re
import socket
from functools import lru_cache

from .directives import (
    _FORK,
    _GET_EXECUTABLE,
    _MEMORY,
    _NGPU,
    _NP,
    _NRANKS,
    _OMP_NUM_THREADS,
    _PROCESSOR_FRACTION,
    _WALLTIME,
    _Directives,
)
from .errors import NoSchedulerError
from .scheduling.base import JobStatus
from .scheduling.fake_scheduler import FakeScheduler
from .scheduling.lsf import LSFScheduler
from .scheduling.pbs import PBSScheduler
from .scheduling.simple_scheduler import SimpleScheduler
from .scheduling.slurm import SlurmScheduler

logger = logging.getLogger(__name__)


@lru_cache(maxsize=1)
def _cached_fqdn():
    """Return the fully qualified domain name.

    This value is cached because fetching the fully qualified domain name can
    be slow on macOS.
    """
    return socket.getfqdn()


class _ComputeEnvironmentType(type):
    """Metaclass used for :class:`~.ComputeEnvironment`.

    This metaclass automatically registers :class:`~.ComputeEnvironment`
    definitions, which enables the automatic determination of the present
    environment. The registry can be obtained from
    :func:`~.registered_environments`.
    """

    def __init__(cls, name, bases, dct):
        if not hasattr(cls, "registry"):
            cls.registry = {}
        else:
            cls.registry[name] = cls
        super().__init__(name, bases, dct)


def template_filter(func):
    """Decorate a function as a :class:`~.ComputeEnvironment` template filter.

    This decorator is applied to methods defined in a subclass of
    :class:`~.ComputeEnvironment` that are used in that environment's
    templates. The decorated function becomes a class method that is available
    as a :ref:`jinja2 filter <jinja2:filters>` in templates rendered by a
    :class:`~.FlowProject` with that :class:`~.ComputeEnvironment`.

    Parameters
    ----------
    func : callable
        Function to decorate.

    Returns
    -------
    callable
        Decorated function.

    """
    setattr(func, "_flow_template_filter", True)
    return classmethod(func)


class ComputeEnvironment(metaclass=_ComputeEnvironmentType):
    """Define computational environments.

    The ComputeEnvironment class allows us to automatically determine
    specific environments in order to programmatically adjust workflows
    in different environments.

    The default method for the detection of a specific environment is to
    provide a regular expression matching the environment's hostname.
    For example, if the hostname is ``my-server.com``, one could identify the
    environment by setting the ``hostname_pattern`` to ``'my-server'``.
    """

    scheduler_type = None
    hostname_pattern = None
    submit_flags = None
    template = "base_script.sh"
    mpi_cmd = "mpiexec"

    @classmethod
    def is_present(cls):
        """Determine whether this specific compute environment is present.

        The default method for environment detection is trying to match a
        hostname pattern or delegate the environment check to the associated
        scheduler type.
        """
        if cls.hostname_pattern is None:
            if cls.scheduler_type is None:
                return False
            return cls.scheduler_type.is_present()
        return re.match(cls.hostname_pattern, _cached_fqdn()) is not None

    @classmethod
    def get_scheduler(cls):
        """Return an environment-specific scheduler driver.

        The returned scheduler class provides a standardized interface to
        different scheduler implementations.
        """
        try:
            return getattr(cls, "scheduler_type")()
        except (AttributeError, TypeError):
            raise NoSchedulerError(
                f"No scheduler defined for environment '{cls.__name__}'."
            )

    @classmethod
    def submit(cls, script, flags=None, *args, **kwargs):
        r"""Submit a job submission script to the environment's scheduler.

        Scripts should be submitted to the environment, instead of directly
        to the scheduler to allow for environment specific post-processing.

        Parameters
        ----------
        script : str
            The script to submit.
        flags : list
            A list of additional flags to provide to the scheduler.
            (Default value = None)
        \*args
            Positional arguments forwarded to the scheduler's submit method.
        \*\*kwargs
            Keyword arguments forwarded to the scheduler's submit method.

        Returns
        -------
        JobStatus.submitted or None
            Status of job, if submitted.

        """
        if flags is None:
            flags = []
        env_flags = getattr(cls, "submit_flags", [])
        if env_flags:
            flags.extend(env_flags)
        # parse the flag to check for --job-name
        for flag in flags:
            if "--job-name" in flag:
                raise ValueError('Assignment of "--job-name" is not supported.')
        # Hand off the actual submission to the scheduler
        if cls.get_scheduler().submit(script, flags=flags, *args, **kwargs):
            return JobStatus.submitted
        return None

    @classmethod
    def add_args(cls, parser):
        """Add arguments related to this compute environment to an argument parser.

        Parameters
        ----------
        parser : :class:`argparse.ArgumentParser`
            The argument parser where arguments will be added.

        """
        pass

    @classmethod
    def _get_omp_prefix(cls, operation):
        """Get the OpenMP prefix based on the ``omp_num_threads`` directive.

        Parameters
        ----------
        operation : :class:`flow.project._JobOperation`
            The operation to be prefixed.

        Returns
        -------
        str
            The prefix to be added to the operation's command.

        """
        return "export OMP_NUM_THREADS={}; ".format(
            operation.directives["omp_num_threads"]
        )

    @classmethod
    def _get_mpi_prefix(cls, operation, parallel):
        """Get the MPI prefix based on the ``nranks`` directives.

        Parameters
        ----------
        operation : :class:`flow.project._JobOperation`
            The operation to be prefixed.
        parallel : bool
            If True, operations are assumed to be executed in parallel, which
            means that the number of total tasks is the sum of all tasks
            instead of the maximum number of tasks. Default is set to False.

        Returns
        -------
        str
            The prefix to be added to the operation's command.

        """
        if operation.directives.get("nranks"):
            return "{} -n {} ".format(cls.mpi_cmd, operation.directives["nranks"])
        return ""

    @template_filter
    def get_prefix(cls, operation, parallel=False, mpi_prefix=None, cmd_prefix=None):
        """Template filter generating a command prefix from directives.

        Parameters
        ----------
        operation : :class:`flow.project._JobOperation`
            The operation to be prefixed.
        parallel : bool
            If True, operations are assumed to be executed in parallel, which means
            that the number of total tasks is the sum of all tasks instead of the
            maximum number of tasks. Default is set to False.
        mpi_prefix : str
            User defined mpi_prefix string. Default is set to None.
            This will be deprecated and removed in the future.
        cmd_prefix : str
            User defined cmd_prefix string. Default is set to None.
            This will be deprecated and removed in the future.

        Returns
        -------
        str
            The prefix to be added to the operation's command.

        """
        prefix = ""
        if operation.directives.get("omp_num_threads"):
            prefix += cls._get_omp_prefix(operation)
        if mpi_prefix:
            prefix += mpi_prefix
        else:
            prefix += cls._get_mpi_prefix(operation, parallel)
        if cmd_prefix:
            prefix += cmd_prefix
        # if cmd_prefix and if mpi_prefix for backwards compatibility
        # Can change to get them from directives for future
        return prefix

    @classmethod
    def _get_default_directives(cls):
        return _Directives(
            (
                _GET_EXECUTABLE(),
                _FORK,
                _MEMORY,
                _NGPU,
                _NP,
                _NRANKS,
                _OMP_NUM_THREADS,
                _PROCESSOR_FRACTION,
                _WALLTIME,
            )
        )


class StandardEnvironment(ComputeEnvironment):
    """Default environment which is always present."""

    @classmethod
    def is_present(cls):
        """Determine whether this specific compute environment is present.

        The StandardEnvironment is always present, so this returns True.
        """
        return True


class TestEnvironment(ComputeEnvironment):
    """Environment used for testing.

    The test environment will print a mocked submission script
    and submission commands to screen. This enables testing of
    the job submission script generation in environments without
    a real scheduler.
    """

    scheduler_type = FakeScheduler


class SimpleSchedulerEnvironment(ComputeEnvironment):
    """An environment for the simple-scheduler scheduler."""

    scheduler_type = SimpleScheduler
    template = "simple_scheduler.sh"


class DefaultPBSEnvironment(ComputeEnvironment):
    """Default environment for clusters with a PBS scheduler."""

    scheduler_type = PBSScheduler
    template = "pbs.sh"

    @classmethod
    def add_args(cls, parser):
        """Add arguments to the parser.

        Parameters
        ----------
        parser : :class:`argparse.ArgumentParser`
            The argument parser where arguments will be added.

        """
        super().add_args(parser)
        parser.add_argument(
            "--hold", action="store_true", help="Submit jobs, but put them on hold."
        )
        parser.add_argument(
            "--after",
            type=str,
            help="Schedule this job to be executed after "
            "completion of a cluster job with this id.",
        )
        parser.add_argument(
            "--job-output",
            type=str,
            help=(
                "What to name the job output file. "
                "If omitted, uses the scheduler default name. "
                "Both stdout and stderr will be combined."
            ),
        )
        parser.add_argument(
            "--no-copy-env",
            action="store_true",
            help="Do not copy current environment variables into compute node environment.",
        )


class DefaultSlurmEnvironment(ComputeEnvironment):
    """Default environment for clusters with a SLURM scheduler."""

    scheduler_type = SlurmScheduler
    template = "slurm.sh"

    @classmethod
    def add_args(cls, parser):
        """Add arguments to the parser.

        Parameters
        ----------
        parser : :class:`argparse.ArgumentParser`
            The argument parser where arguments will be added.

        """
        super().add_args(parser)
        parser.add_argument(
            "--hold", action="store_true", help="Submit jobs, but put them on hold."
        )
        parser.add_argument(
            "--after",
            type=str,
            help="Schedule this job to be executed after "
            "completion of a cluster job with this id.",
        )
        parser.add_argument(
            "--job-output",
            type=str,
            help=(
                "What to name the job output file. "
                "If omitted, uses the scheduler default name. "
                "Both stdout and stderr will be combined."
            ),
        )


class DefaultLSFEnvironment(ComputeEnvironment):
    """Default environment for clusters with a LSF scheduler."""

    scheduler_type = LSFScheduler
    template = "lsf.sh"

    @classmethod
    def add_args(cls, parser):
        """Add arguments to the parser.

        Parameters
        ----------
        parser : :class:`argparse.ArgumentParser`
            The argument parser where arguments will be added.

        """
        super().add_args(parser)
        parser.add_argument(
            "--hold", action="store_true", help="Submit jobs, but put them on hold."
        )
        parser.add_argument(
            "--after",
            type=str,
            help="Schedule this job to be executed after "
            "completion of a cluster job with this id.",
        )
        parser.add_argument(
            "--job-output",
            type=str,
            help=(
                "What to name the job output file. "
                "If omitted, uses the scheduler default name. "
                "Both stdout and stderr will be combined."
            ),
        )


def registered_environments():
    """Return a list of registered environments.

    Returns
    -------
    list
        List of registered environments.

    """
    return list(ComputeEnvironment.registry.values())


def get_environment(test=False):
    """Attempt to detect the present environment.

    This function iterates through all defined :class:`~.ComputeEnvironment`
    classes in reversed order of definition and returns the first
    environment where the :meth:`~.ComputeEnvironment.is_present` method
    returns True.

    Parameters
    ----------
    test : bool
        Whether to return the TestEnvironment. (Default value = False)
    import_configured : bool
        Whether to import environments specified in the flow configuration.
        (Default value = True)

    Returns
    -------
    :class:`~.ComputeEnvironment`
        The detected environment class.

    """
    if test:
        return TestEnvironment

    # Obtain a list of all registered environments
    env_types = registered_environments()
    logger.debug(
        "List of registered environments:\n\t{}".format(
            "\n\t".join(str(env.__name__) for env in env_types)
        )
    )

    # Select environment based on environment variable if set.
    env_from_env_var = os.environ.get("SIGNAC_FLOW_ENVIRONMENT")
    if env_from_env_var:
        for env_type in env_types:
            if env_type.__name__ == env_from_env_var:
                return env_type
        raise ValueError(f"Unknown environment '{env_from_env_var}'.")

    # Select based on DEBUG flag:
    for env_type in env_types:
        if getattr(env_type, "DEBUG", False):
            logger.debug(f"Select environment '{env_type.__name__}'; DEBUG=True.")
            return env_type

    # Default selection:
    for env_type in reversed(env_types):
        if env_type.is_present():
            logger.debug(f"Select environment '{env_type.__name__}'; is present.")
            return env_type

    # Otherwise, just return a standard environment
    return StandardEnvironment



# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Directives define how workflows are executed.

Directives affect both execution and submission, such as specifying the number
of processors required for an operation.
"""
import datetime
import functools
import operator
import sys
from collections.abc import MutableMapping

from flow.errors import DirectivesError


class _Directive:
    """The definition of a single directive.

    Logic for validation of values when setting, defaults, and the ability
    for directives to inspect other directives (such as using ``nranks`` and
    ``omp_num_threads`` for computing ``np``). This is only meant to work
    with the internals of signac-flow.

    The validation of a directive occurs before the call to ``finalize``. It is
    the caller's responsibility to ensure that finalized values are still valid.

    Since directive values can be dependent on jobs we allow all directives to
    be set to a callable which is lazily validated.

    Parameters
    ----------
    name : str
        The name of the directive.
    validator : callable, optional
        A callable that accepts inputs and attempts to convert the input to a
        valid value for the directive. If it fails, it should raise an
        appropriate error. If not provided or ``None``, the validator
        callable directly returns the passed value. Defaults to ``None``.
    default : any, optional
        Sets the default for the directive, defaults to ``None``.
    serial : callable, optional
        A callable that takes two inputs for the directive and returns the
        appropriate value for these operations running in serial. If ``None`` or
        not provided, the ``max`` function is used. Defaults to ``None``.
    parallel : callable, optional
        A callable that takes two inputs for the directive and returns the
        appropriate value for these operations running in parallel. If ``None``
        or not provided, the ``operator.add`` function is used. Defaults to
        ``None``.  Defaults to ``None``.
    finalize : callable, optional
        A callable that takes the set value of the directive and the
        :class:`~._Directives` object it is a child of and outputs the finalized
        value for that directive. This is useful if some directives have
        multiple ways to be set or are dependent in some way on other
        directives. If ``None`` or not provided, the set value is returned.
        Defaults to ``None``.

    """

    def __init__(
        self,
        name,
        *,
        validator=None,
        default=None,
        serial=max,
        parallel=operator.add,
        finalize=None,
    ):
        self._name = name
        self._default = default
        self._serial = serial
        self._parallel = parallel

        def identity(value):
            return value

        def default_finalize(value, directives):
            return value

        self._validator = identity if validator is None else validator
        self._finalize = default_finalize if finalize is None else finalize

    def __call__(self, value):
        """Return a validated value for the given directive.

        Parameters
        ----------
        value : any
            The value to be validated by the :class:`~._Directive`. Accepted
            values depend on the given instance. All `~._Directive`
            instances support lazily validating callables of single (or multiple) jobs
            that return an acceptable value for the given `~._Directive`
            instance.

        Returns
        -------
        any
            Returns a immediately validated value for the given directive, or if
            a callable was passed, a new callable is returned that wraps the
            given callable with a validator.

        """
        if callable(value):

            @functools.wraps(value)
            def validate_callable(*jobs):
                return self._validator(value(*jobs))

            return validate_callable
        return self._validator(value)


class _Directives(MutableMapping):
    """Class that maps environment and user specified objects for execution.

    Resource request and execution style directives are the primary purpose for
    this class. Examples include the number of processes expected to run or the
    time to request to run.

    Parameters
    ----------
    environment_directives : sequence of _Directive
        The sequence of all environment-specified directives. All other
        directives are user-specified and not validated. All environment
        directives must be specified at initialization.

    """

    def __init__(self, environment_directives):
        self._directive_definitions = {}
        self._defined_directives = {}
        self._user_directives = {}
        self._evaluated = False

        for directive in environment_directives:
            self._add_directive(directive)

    def _add_directive(self, directive):
        if not isinstance(directive, _Directive):
            raise TypeError(
                f"Expected a _Directive object. Received {type(directive)}."
            )
        if directive._name in self._directive_definitions:
            raise ValueError(f"Cannot redefine directive name {directive._name}.")
        self._directive_definitions[directive._name] = directive
        self._defined_directives[directive._name] = directive._default

    def _set_defined_directive(self, key, value):
        try:
            self._defined_directives[key] = self._directive_definitions[key](value)
        except (KeyError, ValueError, TypeError) as error:
            raise DirectivesError(f"Error setting directive {key}") from error

    def __getitem__(self, key):
        if key in self._defined_directives and key in self._directive_definitions:
            value = self._defined_directives[key]
            return self._directive_definitions[key]._finalize(value, self)
        if key in self._user_directives:
            return self._user_directives[key]
        raise KeyError(f"{key} not in directives.")

    def __setitem__(self, key, value):
        if key in self._directive_definitions:
            self._set_defined_directive(key, value)
        else:
            self._user_directives[key] = value
        self._evaluated = False

    def __delitem__(self, key):
        if key in self._directive_definitions:
            self._defined_directives[key] = self._directive_definitions[key]._default
        else:
            del self._user_directives[key]

    def __iter__(self):
        yield from self._defined_directives
        yield from self._user_directives

    def __len__(self):
        return len(self._defined_directives) + len(self._user_directives)

    def __str__(self):
        return str(dict(self))

    def __repr__(self):
        return f"_Directives({str(self)})"

    def update(self, other, aggregate=False, jobs=None, parallel=False):
        """Update directives with another set of directives.

        This method accounts for serial/parallel behavior and aggregation.

        Parameters
        ----------
        other : :class:`~._Directives`
            The other set of directives.
        aggregate : bool
            Whether to combine directives according to serial/parallel rules.
        jobs : :class:`signac.contrib.job.Job` or tuple of :class:`signac.contrib.job.Job`
            The jobs used to evaluate directives.
        parallel : bool
            Whether to aggregate according to parallel rules.

        """
        if aggregate:
            self._aggregate(other, jobs=jobs, parallel=parallel)
        else:
            super().update(other)

    def evaluate(self, jobs):
        """Evaluate directives for the provided jobs.

        This method updates the directives in place, replacing callable
        directives with their evaluated values.

        Parameters
        ----------
        jobs : :class:`signac.contrib.job.Job` or tuple of :class:`signac.contrib.job.Job`
            The jobs used to evaluate directives.

        """
        if not self._evaluated:
            for key, value in self.items():
                self[key] = _evaluate(value, jobs)
            self._evaluated = True

    def _aggregate(self, other, jobs=None, parallel=False):
        self.evaluate(jobs)
        other.evaluate(jobs)
        agg_func_attr = "_parallel" if parallel else "_serial"
        for name in self._defined_directives:
            agg_func = getattr(self._directive_definitions[name], agg_func_attr)
            default_value = self._directive_definitions[name]._default
            other_directive = other.get(name, default_value)
            directive = self[name]
            if other_directive is not None:
                self._defined_directives[name] = agg_func(directive, other_directive)

    @property
    def user_keys(self):  # noqa: D401
        """A generator of user specified keys."""
        return (key for key in self._user_directives)


def _evaluate(value, jobs):
    if callable(value):
        if jobs is None:
            raise RuntimeError(
                "jobs must be specified when evaluating a callable directive."
            )
        return value(*jobs)
    return value


class _OnlyTypes:
    def __init__(self, *types, preprocess=None, postprocess=None):
        def identity(value):
            return value

        self.types = types
        self.preprocess = identity if preprocess is None else preprocess
        self.postprocess = identity if postprocess is None else postprocess

    def __call__(self, value):
        return self.postprocess(self._validate(self.preprocess(value)))

    def _validate(self, value):
        if isinstance(value, self.types):
            return value
        for type_ in self.types:
            try:
                return type_(value)
            except Exception:
                pass
        raise TypeError(
            f"Expected an object convertible to one of the following types: {self.types}. "
            f"Received object {value} of type {type(value)}."
        )


def _raise_below(threshold, allow_none=False):
    def is_greater_or_equal(value):
        if allow_none and value is None:
            return value
        try:
            if value < threshold:
                raise ValueError
        except (TypeError, ValueError):
            raise ValueError(
                f"Expected a number greater than or equal to {threshold}. Received {value}."
            )
        return value

    return is_greater_or_equal


_NP_DEFAULT = 1


def _finalize_np(np, directives):
    """Return the actual number of processes/threads to use.

    We check the default np because when aggregation occurs we multiply the
    number of MPI ranks and OMP_NUM_THREADS. If we always took the greater of
    the given NP and ranks * threads then after aggregating we will inflate the
    number of processors needed as (r1 * t1) + (r2 * t2) <= (r1 + r2) * (t1 + t2)
    for numbers greater than one.
    """
    if callable(np) or np != _NP_DEFAULT:
        return np
    nranks = directives.get("nranks", 1)
    omp_num_threads = directives.get("omp_num_threads", 1)
    if callable(nranks) or callable(omp_num_threads):
        return np
    return max(np, max(1, nranks) * max(1, omp_num_threads))


# Helper validators for defining _Directive
def _no_aggregation(value, other):
    """Return the first argument.

    This is used for directives that ignore aggregation rules.
    """
    return value


def _is_fraction(value):
    if 0 <= value <= 1:
        return value
    raise ValueError("Value must be between 0 and 1.")


def _parse_walltime(walltime):
    """Parse walltime from a walltime directive passed by a user.

    A valid walltime argument is defined as:

    1. Numeric value indicating walltime requested in hours.
    2. :class:`datetime.timedelta` value.
    3. None to indicate no specific walltime request.

    Parameters
    ----------
    walltime : float or :class:`datetime.timedelta` or None
        Requested walltime.

    Returns
    -------
    :class:`datetime.timedelta` or None
        The parsed walltime value.
    """
    if walltime is None:
        return None
    if not isinstance(walltime, datetime.timedelta):
        walltime = datetime.timedelta(hours=walltime)
    return walltime


def _parse_memory(memory):
    """Parse memory value from a memory directive passed by a user.

    A valid memory argument is defined as:

    1. Numeric value with suffix "G" or "GB" indicating memory requested in gigabytes.
    2. Numeric value with suffix "M" or "MB" indicating memory requested in megabytes.
    3. Numeric value with no suffix indicating memory requested in gigabytes.
    4. None to indicate no specific memory request.

    Memory arguments passed as strings are case insensitive.

    Parameters
    ----------
    memory : str or float or None
        Requested memory.

    Returns
    -------
    float or None
        The parsed memory value in gigabytes.
    """
    try:
        if memory is None:
            return None
        # Convert to uppercase, remove "B" suffix if provided
        memory = str(memory).upper().rstrip("B")
        size_type = memory[-1]
        if size_type == "M":
            return float(memory[:-1]) / 1024
        elif size_type == "G":
            return float(memory[:-1])
        else:
            return float(memory)
    except ValueError:
        raise ValueError(
            'Invalid memory passed. For gigabytes use suffix "G" or "GB", '
            'for megabytes use suffix "M" or "MB". If a numeric value is passed then '
            "it will be interpreted as memory in gigabytes."
        )


def _max_not_none(value, other):
    """Return the max of two values, with special handling of None.

    This is used for memory directives in serial and walltime directives in
    parallel.
    """
    if value is None and other is None:
        return None
    elif other is None:
        return value
    elif value is None:
        return other
    else:
        return max(value, other)


def _sum_not_none(value, other):
    """Return the sum of two values, with special handling of None.

    This is used for memory directives in parallel and walltime directives in
    serial.
    """
    if value is None and other is None:
        return None
    elif other is None:
        return value
    elif value is None:
        return other
    else:
        return operator.add(value, other)


# Definitions used for validating directives
_bool = _OnlyTypes(bool)
_natural_number = _OnlyTypes(int, postprocess=_raise_below(1))
_nonnegative_int = _OnlyTypes(int, postprocess=_raise_below(0))
_positive_real_walltime = _OnlyTypes(
    float,
    datetime.timedelta,
    type(None),
    preprocess=_parse_walltime,
    postprocess=_raise_below(datetime.timedelta(seconds=1), True),
    # 1 second is an arbitrarily chosen minimum threshold
)
_positive_real_memory = _OnlyTypes(
    float,
    str,
    type(None),
    preprocess=_parse_memory,
    postprocess=_raise_below(1e-12, True),
    # 1e-12 is an arbitrarily chosen minimum threshold for what constitutes 0
)


# Common directives and their instantiation as _Directive
def _GET_EXECUTABLE():
    # Evaluate the executable directive at call-time instead of definition time.
    # This is because we mock `sys.executable` while generating template reference data.
    _EXECUTABLE = _Directive(
        "executable",
        validator=_OnlyTypes(str),
        default=sys.executable,
        serial=_no_aggregation,
        parallel=_no_aggregation,
    )
    _EXECUTABLE.__doc__ = """Return the path to the executable to be used for an operation.

The executable directive expects a string pointing to a valid executable
file in the current file system.

When called, by default this should point to a Python executable (interpreter);
however, if the :class:`FlowProject` path is an empty string, the executable
can be a path to an executable Python script. Defaults to ``sys.executable``.
"""
    return _EXECUTABLE


_FORK = _Directive("fork", validator=_bool, default=False)
_FORK.__doc__ = """The fork directive can be set to True to enforce that a
particular operation is always executed within a subprocess and not within the
Python interpreter's process even if there are no other reasons that would prevent that.

.. note::

    Setting ``fork=False`` will not prevent forking if there are other reasons for forking,
    such as a timeout.
"""

_MEMORY = _Directive(
    "memory",
    validator=_positive_real_memory,
    default=None,
    serial=_max_not_none,
    parallel=_sum_not_none,
)
_MEMORY.__doc__ = """The memory to request for this operation.

The memory to validate should be either a float, int, or string.
A valid memory argument is defined as:

- Positive numeric value with suffix "g" or "G" indicating memory requested in gigabytes.

For example:

.. code-block:: python

    @Project.operation.with_directives({"memory": "4g"})
    def op(job):
        pass

- Positive numeric value with suffix "m" or "M" indicating memory requested in megabytes.

For example:

.. code-block:: python

    @Project.operation.with_directives({"memory": "512m"})
    def op(job):
        pass

- Positive numeric value with no suffix indicating memory requested in gigabytes.

For example:

.. code-block:: python

    @Project.operation.with_directives({"memory": "4"})
    def op1(job):
        pass

    @Project.operation.with_directives({"memory": 4})
    def op2(job):
        pass
"""

_NGPU = _Directive("ngpu", validator=_nonnegative_int, default=0)
_NGPU.__doc__ = """The number of GPUs to use for this operation.

Expects a nonnegative integer. Defaults to 0.
"""

_NP = _Directive(
    "np", validator=_natural_number, default=_NP_DEFAULT, finalize=_finalize_np
)
_NP.__doc__ = """The number of tasks to launch for a given operation i.e., the number of CPU
cores to be requested for a given operation.

Expects a natural number (i.e. an integer >= 1). This directive introspects into
the "nranks" or "omp_num_threads" directives and uses their product if it is
greater than the current set value. Defaults to 1.
"""

_NRANKS = _Directive("nranks", validator=_nonnegative_int, default=0)
_NRANKS.__doc__ = """The number of MPI ranks to use for this operation. Defaults to 0.

Expects a nonnegative integer.
"""

_OMP_NUM_THREADS = _Directive("omp_num_threads", validator=_nonnegative_int, default=0)
_OMP_NUM_THREADS.__doc__ = """The number of OpenMP threads to use for this operation. Defaults to 0.

Expects a nonnegative integer.
"""

_PROCESSOR_FRACTION = _Directive(
    "processor_fraction",
    validator=_OnlyTypes(float, postprocess=_is_fraction),
    default=1.0,
    serial=_no_aggregation,
    parallel=_no_aggregation,
)
_PROCESSOR_FRACTION.__doc__ = """Fraction of a resource to use on a single operation.

If set to 0.5 for a bundled job with 20 operations (all with 'np' set to 1), 10
CPUs will be used. Defaults to 1.

.. note::

    This can be particularly useful on Stampede2's launcher.
"""

_WALLTIME = _Directive(
    "walltime",
    validator=_positive_real_walltime,
    default=None,
    serial=_sum_not_none,
    parallel=_max_not_none,
)
_WALLTIME.__doc__ = """The number of hours to request for executing this job.

This directive expects a float representing the walltime in hours. Fractional
values are supported. For example, a value of 0.5 will request 30 minutes of
walltime. If no walltimes are requested, the submission will not specify a
walltime in the output script. Some schedulers have a default value that will
be used.

For example:

.. code-block:: python

    @Project.operation.with_directives({"walltime": 24})
    def op(job):
        # This operation takes 1 day to run
        pass

"""


def _document_directive(directive):
    """Dynamically generates documentation for a directive."""
    name = directive._name
    name = name.replace("_", r"\_")
    doc = directive.__doc__
    return f"**{name}**\n\n{doc}"



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Workflow management based on the signac framework.

The signac-flow package provides the basic infrastructure to easily
configure and implement a workflow to operate on a signac_ data space.

.. _signac: https://signac.io/
"""
from . import environment, environments, errors, hooks, scheduling, testing
from .aggregates import aggregator, get_aggregate_id
from .environment import get_environment
from .operations import cmd, directives, with_job
from .project import FlowProject, IgnoreConditions, classlabel, label, staticlabel
from .template import init

# Import packaged environments unless disabled in config:
from .util.misc import redirect_log
from .version import __version__

__all__ = [
    "environment",
    "environments",
    "errors",
    "hooks",
    "scheduling",
    "testing",
    "aggregator",
    "get_aggregate_id",
    "get_environment",
    "cmd",
    "directives",
    "with_job",
    "FlowProject",
    "IgnoreConditions",
    "classlabel",
    "label",
    "staticlabel",
    "init",
    "redirect_log",
    "__version__",
]



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Definitions of Exception classes used in this package."""

import jinja2
from jinja2.ext import Extension as Jinja2Extension


class ConfigKeyError(KeyError):
    """Indicates that a config key was not found."""

    pass


class DirectivesError(ValueError):
    """Indicates that a directive was incorrectly set."""

    pass


class SubmitError(RuntimeError):
    """Indicates an error during cluster job submission."""

    pass


class NoSchedulerError(AttributeError):
    """Indicates that there is no scheduler type defined for an environment class."""

    pass


class UserConditionError(RuntimeError):
    """Indicates an error during evaluation of a condition."""

    pass


class UserOperationError(RuntimeError):
    """Indicates an error during execution of a :class:`~.BaseFlowOperation`."""

    pass


class TemplateError(Jinja2Extension):
    """Indicates an error in a jinja2 template."""

    # Reference: https://jinja.palletsprojects.com/en/2.11.x/extensions/
    # The tags are a set of strings that trigger the parse method.
    tags = {"raise"}

    def parse(self, parser):
        """Call :meth:`~.err` when a template raises an Exception."""
        lineno = next(parser.stream).lineno
        args = [parser.parse_expression()]
        return jinja2.nodes.CallBlock(
            self.call_method("err", args), [], [], []
        ).set_lineno(lineno)

    def err(self, msg, caller):
        """Raise a template error."""
        raise jinja2.TemplateError(msg)


class FlowProjectDefinitionError(ValueError):
    """Indicates an invalid FlowProject definition."""

    pass



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Initialize a project for testing."""
import signac

from .template import init


def make_project(alias="project", root=None, **kwargs):
    r"""Initialize a project for testing.

    The initialized project has a few operations and a few jobs that are in
    various points in the workflow defined by the project.

    Parameters
    ----------
    alias : str
        Python identifier used as a file name for the template output.
        (Default value = ``"project"``)
    root : str
        Directory where the output file is placed. Uses the current working
        directory if None. (Default value = None)
    \*\*kwargs
        Keyword arguments forwarded to :meth:`signac.testing.init_jobs`.

    Returns
    -------
    :class:`signac.Project`
        Project with initialized jobs.

    """
    init(alias=alias, root=root, template="testing")
    project = signac.init_project(name=alias, root=root)
    signac.testing.init_jobs(project, **kwargs)
    return project



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Operation hooks."""
import logging

logger = logging.getLogger(__name__)


class _HooksList(list):
    def __call__(self, *args, **kwargs):
        """Call all hook functions as part of this list."""
        for hook in self:
            logger.debug(f"Executing hook function '{hook}'.")
            try:
                hook(*args, **kwargs)
            except Exception as error:
                logger.error(
                    "Error occurred during execution of "
                    "hook '{}': {}.".format(hook, error)
                )
                raise


class _Hooks:
    """:class:`~._Hooks` execute an action or set of actions at specific stages of operation execution.

    :class:`~._Hooks` can execute a user defined function when an operation
    starts, succeeds, fails, or finishes (regardless of whether the operation
    executed successfully or failed).

    Hooks can be installed at the operation level as decorators, or on an
    instance of :class:`~.FlowProject` through
    :meth:`~.FlowProject.project_hooks`.

    Examples
    --------
    The example below shows an operation level decorator that prints the
    operation name and job id at the start of the operation execution.

    .. code-block:: python

        def start_hook(operation_name, job):
            print(f"Starting operation {operation_name} on job {job.id}.")

        @FlowProject.operation
        @FlowProject.operation_hooks.on_start(start_hook)
        def foo(job):
            pass

    Parameters
    ----------
    on_start : list of callables
        Function(s) to execute before the operation begins execution.
    on_exit : list of callables
        Function(s) to execute after the operation exits, with or without errors.
    on_success : list of callables
        Function(s) to execute after the operation exits without error.
    on_exception : list of callables
        Function(s) to execute after the operation exits with an error.

    """

    _hook_triggers = [
        "on_start",
        "on_exit",
        "on_success",
        "on_exception",
    ]

    def __init__(
        self, *, on_start=None, on_exit=None, on_success=None, on_exception=None
    ):
        def set_hooks(self, trigger_name, trigger_value):
            if trigger_value is None:
                trigger_value = []
            setattr(self, trigger_name, _HooksList(trigger_value))

        set_hooks(self, "on_start", on_start)
        set_hooks(self, "on_exit", on_exit)
        set_hooks(self, "on_success", on_success)
        set_hooks(self, "on_exception", on_exception)

    def __setattr__(self, name, value):
        """Convert to _HooksList when setting a hook trigger attribute."""
        if name in self._hook_triggers:
            super().__setattr__(name, _HooksList(value))
        else:
            super().__setattr__(name, value)

    def update(self, other):
        """Update this instance with hooks from another instance."""
        for hook_trigger in self._hook_triggers:
            getattr(self, hook_trigger).extend(getattr(other, hook_trigger))

    def __str__(self):
        """Return a string representation of all hooks."""
        return "{}({})".format(
            type(self).__name__,
            ", ".join(
                f"{hook_trigger}={getattr(self, hook_trigger)}"
                for hook_trigger in self._hook_triggers
            ),
        )

    def __bool__(self):
        """Return True if hooks are defined."""
        return any(
            getattr(self, hook_trigger, None) for hook_trigger in self._hook_triggers
        )



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Defines operation decorators and a simple command line interface ``run``.

This module implements the run() function, which when called equips a regular
Python module with a command line interface. This interface can be used to
execute functions defined within the same module that operate on a signac data
space.

See also: :class:`~.FlowProject`.
"""
import inspect
import logging
from functools import wraps
from textwrap import indent

from .directives import _document_directive
from .environment import ComputeEnvironment
from .errors import FlowProjectDefinitionError

logger = logging.getLogger(__name__)


def cmd(func):
    """Indicate that ``func`` returns a shell command with this decorator.

    If this function is an operation function defined by :class:`~.FlowProject`, it will
    be interpreted to return a shell command, instead of executing the function itself.

    For example:

    .. code-block:: python

        @FlowProject.operation
        @flow.cmd
        def hello(job):
            return "echo {job.id}"

    .. note::
        The final shell command generated for :meth:`~.FlowProject.run` or
        :meth:`~.FlowProject.submit` still respects directives and will prepend e.g. MPI or OpenMP
        prefixes to the shell command provided here.
    """
    if getattr(func, "_flow_with_job", False):
        raise FlowProjectDefinitionError(
            "The @flow.cmd decorator must appear below the @flow.with_job decorator."
        )
    setattr(func, "_flow_cmd", True)
    return func


def with_job(func):
    """Use ``arg`` as a context manager for ``func(arg)`` with this decorator.

    This decorator can only be used for operations that accept a single job as a parameter.

    If this function is an operation function defined by :class:`~.FlowProject`, it will
    be the same as using ``with job:``.

    For example:

    .. code-block:: python

        @FlowProject.operation
        @flow.with_job
        def hello(job):
            print("hello {}".format(job))

    Is equivalent to:

    .. code-block:: python

        @FlowProject.operation
        def hello(job):
            with job:
                print("hello {}".format(job))

    This also works with the `@cmd` decorator:

    .. code-block:: python

        @FlowProject.operation
        @with_job
        @cmd
        def hello(job):
            return "echo 'hello {}'".format(job)

    Is equivalent to:

    .. code-block:: python

        @FlowProject.operation
        @cmd
        def hello_cmd(job):
            return 'trap "cd `pwd`" EXIT && cd {} && echo "hello {job}"'.format(job.ws)
    """
    if getattr(func, "_flow_aggregate", False):
        raise FlowProjectDefinitionError(
            "The @with_job decorator cannot be used with aggregation."
        )

    @wraps(func)
    def decorated(*jobs):
        with jobs[0] as job:
            if getattr(func, "_flow_cmd", False):
                return f'trap "cd $(pwd)" EXIT && cd {job.ws} && {func(job)}'
            else:
                return func(job)

    setattr(decorated, "_flow_with_job", True)
    return decorated


class directives:
    """Decorator for operation functions to provide additional execution directives.

    Directives can for example be used to provide information about required resources
    such as the number of processes required for execution of parallelized operations.
    For more information, read about :ref:`signac-docs:cluster_submission_directives`.

    .. deprecated:: 0.15
        This decorator is deprecated and will be removed in 1.0.
        Use :class:`FlowProject.operation.with_directives` instead.

    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    @classmethod
    def copy_from(cls, func):
        """Copy directives from another operation."""
        return cls(**getattr(func, "_flow_directives", {}))

    def __call__(self, func):
        """Add directives to the function.

        This call operator allows the class to be used as a decorator.

        Parameters
        ----------
        func : callable
            The function to decorate.

        Returns
        -------
        callable
            The decorated function.

        """
        directives = getattr(func, "_flow_directives", {})
        directives.update(self.kwargs)
        setattr(func, "_flow_directives", directives)
        return func


# Remove when @flow.directives is removed
_directives_to_document = (
    ComputeEnvironment._get_default_directives()._directive_definitions.values()
)
directives.__doc__ += indent(
    "\n**Supported Directives:**\n\n"
    + "\n\n".join(
        _document_directive(directive) for directive in _directives_to_document
    ),
    "    ",
)


def _get_operations(include_private=False):
    """Yield the name of all functions that qualify as an operation function.

    The module is inspected and all functions that have only one argument
    is yielded. Unless the 'include_private' argument is True, all private
    functions, that means the name starts with one or more '_' characters
    are ignored.
    """
    module = inspect.getmodule(inspect.currentframe().f_back.f_back)
    for name, obj in inspect.getmembers(module):
        if not include_private and name.startswith("_"):
            continue
        if inspect.isfunction(obj):
            signature = inspect.getfullargspec(obj)
            if len(signature.args) == 1:
                yield name


__all__ = ["cmd", "directives", "with_job"]



# Copyright (c) 2021 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Drexel University HPC Environments."""
from ..environment import DefaultSlurmEnvironment


class PicotteEnvironment(DefaultSlurmEnvironment):
    """Environment profile for the Picotte supercomputer.

    https://proteusmaster.urcf.drexel.edu/urcfwiki/
    Note: This link may require access to the Drexel University network.
    """

    hostname_pattern = r".*\.cm\.cluster$"
    template = "drexel-picotte.sh"

    @classmethod
    def add_args(cls, parser):
        """Add arguments to parser.

        Parameters
        ----------
        parser : :class:`argparse.ArgumentParser`
            The argument parser where arguments will be added.

        """
        super().add_args(parser)
        parser.add_argument(
            "--partition",
            choices=("def", "gpu"),
            default="def",
            help="Specify the partition to submit to. (default=def)",
        )


__all__ = ["PicotteEnvironment"]



# Copyright (c) 2019 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Define the signac-flow version."""

__version__ = "0.21.0"

__all__ = ["__version__"]



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""The environments module contains a set of provided environment profiles."""
from . import drexel, incite, umich, umn, xsede

__all__ = [
    "drexel",
    "incite",
    "umich",
    "umn",
    "xsede",
]



# Copyright (c) 2016 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import os

from setuptools import find_packages, setup

requirements = [
    # The core package.
    "signac>=1.3.0",
    # For the templated generation of (submission) scripts.
    "jinja2>=3.0.0",
    # To enable the parallelized execution of operations across processes.
    "cloudpickle>=1.6.0",
    # Deprecation management
    "deprecation>=2.0.0",
    # Progress bars
    "tqdm>=4.60.0",
    # For schema validation
    "jsonschema>=3.0.0",
]

description = "Simple workflow management for signac projects."

try:
    this_path = os.path.dirname(os.path.abspath(__file__))
    fn_readme = os.path.join(this_path, "README.md")
    with open(fn_readme) as fh:
        long_description = fh.read()
except OSError:
    long_description = description

setup(
    name="signac-flow",
    version="0.21.0",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=True,
    maintainer="signac Developers",
    maintainer_email="signac-support@umich.edu",
    author="Carl Simon Adorf et al.",
    author_email="csadorf@umich.edu",
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://signac.io",
    download_url="https://pypi.org/project/signac-flow/",
    keywords="workflow management signac framework database",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Topic :: Scientific/Engineering :: Physics",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    entry_points={
        "console_scripts": [
            "flow = flow.__main__:main",
        ],
    },
    scripts=[
        "bin/simple-scheduler",
    ],
    install_requires=requirements,
    # Supported versions are determined according to NEP 29.
    # https://numpy.org/neps/nep-0029-deprecation_policy.html
    python_requires=">=3.8",
)
